/**
 *  Marching Cubes Demo, by glUser3f [glUser3f@gamedev.net]
 *  Feel free to use this code in your own projects.
 *  If you do something cool with it, please email me so I can see it!
 *
 *  Credits:
 *  Theory by Paul Bourke        [http://astronomy.swin.edu.au/~pbourke/modelling/polygonise/]
 *  OpenGL basecode by NeHe      [nehe.gamedev.net]
 *  Lookup Tables by Paul Bourke [http://astronomy.swin.edu.au/~pbourke/modelling/polygonise/]
 *  The rest is done by me, glUser3f ;)
 *
 */



#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <math.h>
#include <tchar.h>

#include "glfw.h"
#include "Grid.h"



#define POW4(X) ((X)*(X)*(X)*(X))



float AmbLight[ 4 ] = { 0.3f, 0.3f,  0.5f, 0.3f };
float DifLight[ 4 ] = { 0.6f, 0.6f,  0.6f, 0.6f };
float LightPos[ 4 ] = { 0.0f, 0.0f, 12.0f, 0.0f };

bool gWire   = false;
bool gLight  = true;
bool gPaused = false;

float zoom = 2.2f;

bool g_bPriorDown = false;
bool g_bNextDown  = false;


BOOL glfwActive = TRUE;
HDC hDC = NULL;


float xrot = 30.0f;
float yrot = 30.0f;


struct Cube
{
	float size;
	float center[ 3 ];
};


float cube( float x, float y, float z );
Grid gGrid(30, 1.0f, cube );


Cube gCubes[ 3 ] =
{
	{ 0.2f , { 0.0f, 0.0f, 0.0f } },
	{ 0.25f, { 0.0f, 0.0f, 0.0f } },
	{ 0.3f,  { 0.0f, 0.0f, 0.0f } }
};


HGLRC     hRC  = NULL;



// ***************************************************************************************
float cube(float x, float y, float z)
{
	float val = 0.0f;
	for (int i = 0; i < 3; i++) {
		val += POW4(gCubes[i].size) / (POW4(x - gCubes[i].center[0]) + POW4(y - gCubes[i].center[1]) + POW4(z - gCubes[i].center[2]));
	}
	return val;
}



// ***************************************************************************************
BOOL glfwInit()
{
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glShadeModel(GL_SMOOTH);
	glLightfv(GL_LIGHT1, GL_AMBIENT, AmbLight);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, DifLight);
	glLightfv(GL_LIGHT1, GL_POSITION, LightPos);
	glEnable(GL_LIGHT1);
	glEnable(GL_COLOR_MATERIAL);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glEnable(GL_LIGHTING);
	return TRUE;
}



// ***************************************************************************************
GLvoid glfwRender()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -zoom);
	glRotatef(xrot, 1.0f, 0.0f, 0.0f);
	glRotatef(yrot, 0.0f, 1.0f, 0.0f);
	gGrid.draw();
	glFlush();
}



// ***************************************************************************************
GLvoid glfwUpdate( const double fTimePassed )
{
	gGrid.calculateGrids();
	gGrid.go();

	static double i = 0.0;

	double sin120 = sin(i/120.0);
	double sin80 = sin(i/80.0);

	gCubes[0].center[0] = sin(i/140.0) * 0.2f;
	gCubes[0].center[1] = sin120       * 0.4f;
	gCubes[0].center[2] = sin(i/100.0) * 0.3f;

	gCubes[1].center[0] = sin(i/40.0)  * 0.7f;
	gCubes[1].center[1] = sin80        * 0.5f;
	gCubes[1].center[2] = sin120       * 0.3f;

	gCubes[2].center[0] = sin80        * 0.6f;
	gCubes[2].center[1] = sin(i/60.0)  * 0.1f;
	gCubes[2].center[2] = sin(i/30.0)  * 0.5f;

	if (!gPaused) {
		i += fTimePassed / 12.5;	// +1 at 80fps
	}

	xrot += 0.032 * fTimePassed;	// +0.4
	yrot += 0.024 * fTimePassed;	// +0.3

	if( g_bPriorDown )
	{
		zoom -= 0.1f;
		if( zoom < 1.3f )
		{
			zoom = 1.3f;
		}
	}
	else if( g_bNextDown )
	{
		zoom += 0.1f;
		if( zoom > 30.0f )
		{
			zoom = 30.0f;
		}
	}
}



// ***************************************************************************************
void glfwResizeScene( GLsizei width, GLsizei height) {
	if (height == 0) {
		height = 1;
	}
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}



// ***************************************************************************************
void glfwKillWnd( const HWND hwnd )
{
	if (hRC) {
		if (!wglMakeCurrent(NULL, NULL)) {
			MessageBox(NULL, TEXT( "Error: Function wglMakeCurrent." ), TEXT( "SHUTDOWN ERROR" ), MB_OK | MB_ICONERROR);
		}
		if (!wglDeleteContext(hRC)) {
			MessageBox(NULL, TEXT( "Error: Function wglDeleteContext." ), TEXT( "SHUTDOWN ERROR" ), MB_OK | MB_ICONERROR);
		}
		hRC = NULL;
	}
}



// ***************************************************************************************
BOOL glfwCreateWnd( const HWND hwnd )
{
	RECT r;
	GetWindowRect( hwnd, &r );
	int width = r.right - r.left;
	int height = r.bottom - r.top;

	GLuint   pixelFormat;

	static	PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		32,							// colordepth
		0, 0, 0, 0, 0, 0,
		0,
		0,
		0,
		0, 0, 0, 0,
		16,
		0,
		0,
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

	if (!(hDC = GetDC(hwnd)))	{
		glfwKillWnd( hwnd );
		MessageBox(NULL, TEXT( "Error: Function GetDC." ), TEXT( "STARTUP ERROR" ), MB_OK | MB_ICONERROR);
		return FALSE;
	}

	if (!(pixelFormat = ChoosePixelFormat(hDC, &pfd)))	{
		glfwKillWnd( hwnd );
		MessageBox(NULL, TEXT( "Error: Function ChoosePixelFormat." ), TEXT( "STARTUP ERROR" ), MB_OK | MB_ICONERROR);
		return FALSE;
	}

	if(!SetPixelFormat(hDC, pixelFormat, &pfd)) {
		glfwKillWnd( hwnd );
		TCHAR sz[ 50 ] = TEXT( "" );
		_stprintf( sz, TEXT( "Error: Function SetPixelFormat. [%u]" ), GetLastError() );
		MessageBox(NULL, sz, TEXT( "STARTUP ERROR" ), MB_OK | MB_ICONERROR);
		return FALSE;
	}

	if (!(hRC = wglCreateContext(hDC))) {
		glfwKillWnd( hwnd );
		MessageBox(NULL, TEXT( "Error: Function wglCreateContext." ), TEXT( "STARTUP ERROR" ), MB_OK | MB_ICONERROR);
		return FALSE;
	}

	if(!wglMakeCurrent(hDC, hRC)) {
		glfwKillWnd( hwnd );
		MessageBox(NULL, TEXT( "Error: Function wglMakeCurrent." ), TEXT( "STARTUP ERROR" ), MB_OK | MB_ICONERROR);
		return FALSE;
	}


	glfwResizeScene(width, height);

	if (!glfwInit()) {
		glfwKillWnd( hwnd );
		MessageBox(NULL, TEXT( "Error: Function glfwInit." ), TEXT( "STARTUP ERROR" ), MB_OK | MB_ICONERROR);
		return FALSE;
	}

	return TRUE;
}

